#ifndef HASH_H
#define HASH_H

typedef struct _hash_link hash_link;
typedef struct _hash_table hash_table;

struct _hash_link {
    unsigned int key;		/* this field saves file descriptor number */
    unsigned int data;		/* MSB 16: interest_set, LSB 16: position in the set */
    hash_link *next;
};

struct _hash_table {
    hash_link **buckets;
    unsigned int current_slot;
    hash_link *next;
    int count;
};

hash_table *hash_create();
void hash_join(hash_table *, hash_link *);
void hash_remove_link(hash_table *, hash_link *);
hash_link *hash_lookup(hash_table *, const int);
void hash_first(hash_table *);
hash_link *hash_next(hash_table *);
void hash_last(hash_table *);
hash_link *hash_get_bucket(hash_table *, unsigned int);
void hashFreeMemory(hash_table *);
void hashFreeItems(hash_table *);

/*
 *  Here are some good prime number choices.  It's important not to
 *  choose a prime number that is too close to exact powers of 2.
 *
 *  HASH_SIZE 103               // prime number < 128
 *  HASH_SIZE 229               // prime number < 256
 *  HASH_SIZE 467               // prime number < 512
 *  HASH_SIZE 977               // prime number < 1024
 *  HASH_SIZE 1979              // prime number < 2048
 *  HASH_SIZE 4019              // prime number < 4096
 *  HASH_SIZE 6037              // prime number < 6144
 *  HASH_SIZE 7951              // prime number < 8192
 *  HASH_SIZE 12149             // prime number < 12288
 *  HASH_SIZE 16231             // prime number < 16384
 *  HASH_SIZE 33493             // prime number < 32768
 *  HASH_SIZE 65357             // prime number < 65536
 */
#define  DEFAULT_HASH_SIZE 7951 /* prime number < 8192 */

#define HASH_FUNC(k, size) (k % size)

#endif
